local date = date
local format = format
local gsub = gsub
local strlen = strlen
local strsub = strsub
local time = time

local GetNumSavedInstances = GetNumSavedInstances
local GetRealmName = GetRealmName
local GetSavedInstanceInfo = GetSavedInstanceInfo
local GetUnitName = GetUnitName
local RequestRaidInfo = RequestRaidInfo
local SecondsToTime = SecondsToTime

local Abacus = Rock("LibAbacus-3.0", false, true)
local Tablet = Rock("Tablet-2.0", false, true)

RaidSaveFu = Rock:NewAddon("RaidSaveFu", "LibRockDB-1.0", "LibRockEvent-1.0", "LibRockConfig-1.0", "LibFuBarPlugin-3.0")
local RaidSaveFu = RaidSaveFu
RaidSaveFu.title = "RaidSaveFu"
RaidSaveFu:SetDatabase("RaidSaveFuDB");

RaidSaveFu:SetDatabaseDefaults('profile', {
	textwhenzero = false,
	shorttext = false,
	showid = true,
	showcooldown = true,
	formatcooldown = "Full",
	showreset = true,
	formatdate = "Chinese",
	dateseparator = "Hyphen",
	formatdayw = "Hide",
	formatdaym = "Show",
	formatmonth = "Numeric",
	formatyear = "Full",
	formattime = "Show24",	
} )
RaidSaveFu:SetFuBarOption('tooltipType', "Tablet-2.0")
RaidSaveFu:SetFuBarOption('configType', "LibRockConfig-1.0")
RaidSaveFu:SetFuBarOption('iconPath', "Interface\\Addons\\FuBar_RaidSaveFu\\icon.tga")
RaidSaveFu:SetFuBarOption('hasNoColor', true)
RaidSaveFu:SetFuBarOption('cannotHideText', false)
RaidSaveFu:SetFuBarOption('clickableTooltip', true)
RaidSaveFu:SetFuBarOption('independentProfile', false)


function RaidSaveFu:OnInitialize()
	local optionsTable = {
		name = "FuBar_RaidSaveFu",
		desc = self.notes,
		handler = RaidSaveFu,
		type = 'group',
		args = {
			instance = {
				type = 'group',
				name = "Instance information",
				desc = "Select instance information to be displayed",
				order = 20,
				args = {
					showID = {
						type = 'toggle',
						name = "Show instance IDs",
						desc = "Show or hide the numeric instance IDs",
						order = 10,
						get = function(value)
							return self.db.profile.showid
						end,
						set = function(value)
							self.db.profile.showid = not self.db.profile.showid
							self:UpdateFuBarTooltip()
							return self.db.profile.showid
						end,
					},
					showCooldown = {
						type = 'toggle',
						name = "Show time remaining",
						desc = "Display the time the instance ID is still valid",
						order = 20,
						get = function(value)
							return self.db.profile.showcooldown
						end,
						set = function(value)
							self.db.profile.showcooldown = not self.db.profile.showcooldown
							self:UpdateFuBarTooltip()
							return self.db.profile.showcooldown
						end,
					},
					formatCooldown = {
						type = 'choice',
						name = "Remaining time format",
						desc = "Display style for the time until reset",
						order = 30,
						get = function(value)
								return self.db.profile.formatcooldown
						end,
						set = function(value)
								self.db.profile.formatcooldown = value
						end,
						choiceType = "dict",
						choices = { 
							Extended = "Extended",
							Full = "Full",
							Short = "Short",
							Condensed = "Condensed",
						},
						choiceDescs = {
							Extended = "Yrs, Mnths, Days, Hrs, Mins, Secs",
							Full = "First two of Yrs, Mnths, Days, Hrs, Mins, Secs",
							Short = "First of Yrs, Mnths, Days, Hrs, Mins, Secs",
							Condensed = "Digits only",
						},
						choiceOrder = { "Extended", "Full", "Short", "Condensed", },
					},
					showReset = {
						type = 'toggle',
						name = "Show reset date and time",
						desc = "Display the date (and/or time) the instance ID expires",
						order = 40,
						get = function(value)
							return self.db.profile.showreset
						end,
						set = function(value)
							self.db.profile.showreset = not self.db.profile.showreset
							self:UpdateFuBarTooltip()
							return self.db.profile.showreset
						end,
					},
					formatDate = {
						type = 'choice',
						name = "Expiry date format",
						desc = "Display style for the date instance IDs expire",
						order = 50,
						get = function(value)
								return self.db.profile.formatdate
						end,
						set = function(value)
								self.db.profile.formatdate = value
						end,
						choiceType = "dict",
						choices = { 
							Chinese = "Chinese",
							USAmerican = "USAmerican",
							European = "European",
						},
						choiceDescs = {
							Chinese = "Day (Week), Year, Month, Day",
							USAmerican = "Day (Week), Month, Day, Year",
							European = "Day (Week), Day, Month, Year",
						},
						choiceOrder = { "Chinese", "USAmerican", "European", },
					},
					dateSeparator = {
						type = 'choice',
						name = "Date separator",
						desc = "Character used in-between Year, Month, Day",
						order = 70,
						get = function(value)
								return self.db.profile.dateseparator
						end,
						set = function(value)
								self.db.profile.dateseparator = value
						end,
						choiceType = "dict",
						choices = { 
							None = "None",
							Space = "Space",
							Hyphen = "Hyphen",
							Slash = "Slash",
						},
						choiceOrder = { "None", "Space", "Hyphen", "Slash", },
					},
					formatDayW = {
						type = 'choice',
						name = "Day of the week style",
						desc = "Style used for displaying the name of the day of the week",
						order = 60,
						get = function(value)
								return self.db.profile.formatdayw
						end,
						set = function(value)
								self.db.profile.formatdayw = value
						end,
						choiceType = "dict",
						choices = { 
							Hide = "Hide",
							Full = "Full name",
							Short = "Short name",
						},
						choiceOrder = { "Hide", "Full", "Short", },
					},
					formatDayM = {
						type = 'choice',
						name = "Day of the month style",
						desc = "Style used for displaying the position of the day in the month",
						order = 80,
						get = function(value)
								return self.db.profile.formatdaym
						end,
						set = function(value)
								self.db.profile.formatdaym = value
						end,
						choiceType = "dict",
						choices = { 
							Hide = "Hide",
							Show = "Show",
						},
						choiceOrder = { "Hide", "Show", },
					},
					formatMonth = {
						type = 'choice',
						name = "Month style",
						desc = "Style used for displaying the month",
						order = 90,
						get = function(value)
								return self.db.profile.formatmonth
						end,
						set = function(value)
								self.db.profile.formatmonth = value
						end,
						choiceType = "dict",
						choices = { 
							Hide = "Hide",
							Numeric = "Numeric",
							Full = "Full name",
							Short = "Short name",
						},
						choiceOrder = { "Hide", "Numeric", "Full", "Short", },
					},
					formatYear = {
						type = 'choice',
						name = "Year style",
						desc = "Style used for displaying the year",
						order = 100,
						get = function(value)
								return self.db.profile.formatyear
						end,
						set = function(value)
								self.db.profile.formatyear = value
						end,
						choiceType = "dict",
						choices = { 
							Hide = "Hide",
							Full = "Full year",
							Short = "No century",
						},
						choiceOrder = { "Hide", "Full", "Short", },
					},
					formatTime = {
						type = 'choice',
						name = "Expiry time format",
						desc = "Display style for the time instance IDs expire",
						order = 110,
						get = function(value)
								return self.db.profile.formattime
						end,
						set = function(value)
								self.db.profile.formattime = value
						end,
						choiceType = "dict",
						choices = { 
							Hide = "Hide",
							Show12 = "12-Hour",
							Show24 = "24-Hour",
						},
						choiceDescs = {
							Hide = "Time not shown",
							Show12 = "H:MM TT",
							Show24 = "HH:MM",
						},
						choiceOrder = { "Hide", "Show12", "Show24", },
					},
				}
			},
			textWhenZero = {
				type = 'toggle',
				name = "Hide text for no saved instances",
				desc = "Show or hide the FuBar text when the character has no saved instances",
				order = 40,
				get = function(value)
					return self.db.profile.textwhenzero
				end,
				set = function(value)
					self.db.profile.textwhenzero = not self.db.profile.textwhenzero
					self:UpdateFuBarText()
					return self.db.profile.textwhenzero
				end,
			},
			shortText = {
				type = 'toggle',
				name = "Short FuBar text",
				desc = "Only show the number of saved instances (no extra text)",
				order = 30,
				get = function(value)
					return self.db.profile.shorttext
				end,
				set = function(value)
					self.db.profile.shorttext = not self.db.profile.shorttext
					self:UpdateFuBarText()
					return self.db.profile.shorttext
				end,
			},
		}
	}
	RaidSaveFu:SetConfigTable(optionsTable)
	RaidSaveFu:SetConfigSlashCommand("/raidsavefu")
	RaidSaveFu.OnMenuRequest = optionsTable
	if self.db.profile.realm == nil then
		self.db.profile.realm = { }
	end
	if not self.db.profile.realm[GetRealmName()] then
		self.db.profile.realm[GetRealmName()] = { }
	end
	if not self.db.profile.realm[GetRealmName()].char then
		self.db.profile.realm[GetRealmName()].char = { }
	end
	if not self.db.profile.realm[GetRealmName()].name then
		self.db.profile.realm[GetRealmName()].name = GetRealmName()
	end
	if not self.db.profile.realm[GetRealmName()].char[GetUnitName("player")] then
		self.db.profile.realm[GetRealmName()].char[GetUnitName("player")] = { }	
	end
	local player = self.db.profile.realm[GetRealmName()].char[GetUnitName("player")]
	if not player.numsaved then
		player.numsaved = 0
	end
	if not player.instance then
		player.instance = { }
	end
	player.name = GetUnitName("player")
	player.class = strupper(UnitClass("player"))
end

function RaidSaveFu:OnEnable()
	self:AddEventListener("Blizzard", "UPDATE_INSTANCE_INFO", "Refresh")
	self:AddEventListener("Blizzard", "RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:AddEventListener("Blizzard", "CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function RaidSaveFu:OnDisable()
	self:RemoveEventListener("Blizzard", "UPDATE_INSTANCE_INFO")
	self:RemoveEventListener("Blizzard", "RAID_INSTANCE_WELCOME")
	self:RemoveEventListener("Blizzard", "CHAT_MSG_SYSTEM")
end

function RaidSaveFu:OnUpdateFuBarText()
	local player = self.db.profile.realm[GetRealmName()].char[GetUnitName("player")]
	local fubartext
	if self.db.profile.shorttext then
		fubartext = format("%d", player.numsaved)
	elseif player.numsaved == 1 then
		fubartext = format("%d Saved Instance", player.numsaved)
	else
		fubartext = format("%d Saved Instances", player.numsaved)
	end
	
	if player.numsaved == 0 and self.db.profile.textwhenzero then
		self:SetFuBarText("");
	else
		self:SetFuBarText(fubartext)
	end
end

function RaidSaveFu:OnUpdateFuBarTooltip()
	local numcolumns = 1
	if self.db.profile.showid then
		numcolumns = numcolumns + 1
	end
	if self.db.profile.showcooldown then
		numcolumns = numcolumns + 1
	end
	if self.db.profile.showreset then
		numcolumns = numcolumns + 1
	end
	local headings = { }
	if self.db.profile.showid then
		headings[1] = HIGHLIGHT_FONT_COLOR_CODE.."Instance ID|r"
		headings[2] = HIGHLIGHT_FONT_COLOR_CODE.."Instance|r"
	else
		headings[1] = HIGHLIGHT_FONT_COLOR_CODE.."Instance|r"
	end
	if self.db.profile.showcooldown then
		if self.db.profile.showid then
			headings[3] = HIGHLIGHT_FONT_COLOR_CODE.."Remaining|r"
		else
			headings[2] = HIGHLIGHT_FONT_COLOR_CODE.."Remaining|r"
		end
	end
	if self.db.profile.showreset then
		if self.db.profile.showid and self.db.profile.showcooldown then
			headings[4] = HIGHLIGHT_FONT_COLOR_CODE.."Expires|r"
		elseif self.db.profile.showid or self.db.profile.showcooldown then
			headings[3] = HIGHLIGHT_FONT_COLOR_CODE.."Expires|r"
		else
			headings[2] = HIGHLIGHT_FONT_COLOR_CODE.."Expires|r"
		end
	end			
	local cat = Tablet:AddCategory(
		'columns', numcolumns
	)
	local emptytooltip = true -- track emptiness to help with inserting space between toons
	local k, v
	for k, v in pairs(self.db.profile.realm[GetRealmName()].char) do
		local toon = v
		if toon.numsaved > 0 then
			local displaytable = { }
			for k, v in pairs(toon.instance) do
				if v.expires > time() then
					local display = { "", "", "", "" }
					local idstring
					if toon.name == GetUnitName("player") then
						idstring = LIGHTYELLOW_FONT_COLOR_CODE..v.id..FONT_COLOR_CODE_CLOSE
					else
						idstring = v.id
					end
					local namestring
					if toon.name == GetUnitName("player") then
						namestring = LIGHTYELLOW_FONT_COLOR_CODE..v.name..FONT_COLOR_CODE_CLOSE
					else
						namestring = v.name
					end
					local cooldownstring = ""
					if self.db.profile.formatcooldown == "Extended" then
						cooldownstring = Abacus:FormatDurationExtended(v.expires - time())
					elseif self.db.profile.formatcooldown == "Condensed" then
						cooldownstring = Abacus:FormatDurationCondensed(v.expires - time())
					elseif self.db.profile.formatcooldown == "Short" then
						cooldownstring = Abacus:FormatDurationShort(v.expires - time())
					else -- Full style
						cooldownstring = SecondsToTime(v.expires - time())
					end
					if v.expires - time() <= 86400 then --less than a day
						cooldownstring = RED_FONT_COLOR_CODE..cooldownstring..FONT_COLOR_CODE_CLOSE
					else -- more than a day
						cooldownstring = GREEN_FONT_COLOR_CODE..cooldownstring..FONT_COLOR_CODE_CLOSE
					end
					local expirestring = self:FormatDate(v.expires)
					if not (self.db.profile.showtime == "Hide") then
						expirestring = expirestring.." "..self:FormatTime(v.expires)
					end
					if toon.name == GetUnitName("player") then
						expirestring = LIGHTYELLOW_FONT_COLOR_CODE..expirestring..FONT_COLOR_CODE_CLOSE
					end
					if self.db.profile.showid then
						display[1] = idstring
						display[2] = namestring
					else
						display[1] = namestring
					end
					if self.db.profile.showcooldown then
						if self.db.profile.showid then
							display[3] = cooldownstring
						else
							display[2] = cooldownstring
						end
					end
					if self.db.profile.showreset then
						if self.db.profile.showid and self.db.profile.showcooldown then
							display[4] = expirestring
						elseif self.db.profile.showid or self.db.profile.showcooldown then
							display[3] = expirestring
						else
							display[2] = expirestring
						end
					end			
					table.insert(displaytable, display)
				end
			end
			if # displaytable > 0 then
				if not emptytooltip then
					cat:AddLine(
						'text', ""
					)
				end
				emptytooltip = false
				cat:AddLine(
					'text', self:ClassColorise(toon.class, toon.name)
				)
				if numcolumns > 1 then
					cat:AddLine(
						'text', headings[1],
						'text2', headings[2],
						'text3', headings[3],
						'text4', headings[4]
					)
					local i
					for i = 1, # displaytable do
						cat:AddLine(
							'text', displaytable[i][1],
							'text2', displaytable[i][2],
							'text3', displaytable[i][3],
							'text4', displaytable[i][4]
						)
					end
				else -- numcolumns == 1
					local i
					for i = 1, # displaytable do
						cat:AddLine(
							'text', displaytable[i][1]
						)
					end
				end
			end
		end
	end
end

function RaidSaveFu:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function RaidSaveFu:RequestRefresh()
	RequestRaidInfo()
end

function RaidSaveFu:Refresh()
	local player = self.db.profile.realm[GetRealmName()].char[GetUnitName("player")]
	player.numsaved = GetNumSavedInstances()
	player.instance = { }
	if player.numsaved == 0 then
	else
		local iname, iid, iexpires, temp
		for i = 1, player.numsaved do
			iname, iid, iexpires = GetSavedInstanceInfo(i)
			iexpires = iexpires + time()
			temp = { name = iname, id = iid, expires = iexpires }
			table.insert(player.instance, temp)
		end
	end
	self:UpdateFuBarPlugin()
end

function RaidSaveFu:GetDaysInMonth(year, month)
	local function is_leap_year(year)
		return year % 4 == 0 and (year % 100 ~= 0 or year % 400 == 0)
	end
	return month == 2 and is_leap_year(year) and 29 or ("\31\28\31\30\31\30\31\31\30\31\30\31"):byte(month)
end

function RaidSaveFu:ClassColorise(class, targetstring)
	local r = format("%x", math.floor(RAID_CLASS_COLORS[gsub(class, " ", "")].r * 255))
	if strlen(r) == 1 then
		r = "0"..r
	end
	local g = format("%x", math.floor(RAID_CLASS_COLORS[gsub(class, " ", "")].g * 255))
	if strlen(g) == 1 then
		g = "0"..g
	end
	local b = format("%x", math.floor(RAID_CLASS_COLORS[gsub(class, " ", "")].b * 255))
	if strlen(b) == 1 then
		b = "0"..b
	end
	return "|cff"..r..g..b..targetstring..FONT_COLOR_CODE_CLOSE
end

function RaidSaveFu:FormatDate(seconds)
	local datestring = ""
	local daym, month, year, sep
	if self.db.profile.formatdayw == "Short" then
		datestring = date("%a", seconds).." "
	elseif self.db.profile.formatdayw == "Full" then
		datestring = date("%A", seconds).." "
	end
	if self.db.profile.dateseparator == "Slash" then
		sep = "/"
	elseif self.db.profile.dateseparator == "Space" then
		sep = " "
	elseif self.db.profile.dateseparator == "Hyphen" then
		sep = "-"
	else -- None
		sep = ""
	end
	if self.db.profile.formatdaym == "Show" then
		daym = date("%d", seconds)..sep
	else -- "Hide"
		daym = ""
	end
	if self.db.profile.formatmonth == "Full" then
		month = date("%B", seconds)..sep
	elseif self.db.profile.formatmonth == "Short" then
		month = date("%b", seconds)..sep
	elseif self.db.profile.formatmonth == "Numeric" then
		month = date("%m", seconds)..sep
	else -- Hide
		month = ""
	end
	if self.db.profile.formatyear == "Full" then
		year = date("%Y", seconds)..sep
	elseif self.db.profile.formatyear == "Short" then
		year = date("%y", seconds)..sep
	else -- Hide
		year = ""
	end
	if self.db.profile.formatdate == "USAmerican" then
		datestring = datestring..month..daym..year
	elseif self.db.profile.formatdate == "European" then
		datestring = datestring..daym..month..year
	else -- Chinese
		datestring = datestring..year..month..daym
	end
	return strsub(datestring, 1, strlen(datestring) - 1)
end

function RaidSaveFu:FormatTime(seconds)
	local timestring = ""
	if self.db.profile.formattime == "Show12" then
		timestring = date("%I", seconds)..":"..date("%M", seconds).." "..date("%p", seconds)
	elseif self.db.profile.formattime == "Show24" then
		timestring = date("%H", seconds)..date("%M", seconds)
	end
	return timestring
end